"""Tests for tuples of all kinds."""

from typing import List, NamedTuple, Tuple

from attrs import Factory, define
from pytest import raises

from cattrs.cols import (
    is_namedtuple,
    namedtuple_dict_structure_factory,
    namedtuple_dict_unstructure_factory,
)
from cattrs.converters import Converter
from cattrs.errors import ForbiddenExtraKeysError


def test_simple_hetero_tuples(genconverter: Converter):
    """Simple heterogenous tuples work.

    Only supported for the Converter (not the BaseConverter).
    """

    genconverter.register_unstructure_hook(int, lambda v: v + 1)

    assert genconverter.unstructure((1, "2"), unstructure_as=Tuple[int, str]) == (
        2,
        "2",
    )

    genconverter.register_structure_hook(int, lambda v, _: v - 1)

    assert genconverter.structure([2, "2"], Tuple[int, str]) == (1, "2")


def test_named_tuple_predicate():
    """The NamedTuple predicate works."""

    assert not is_namedtuple(tuple)
    assert not is_namedtuple(Tuple[int, ...])
    assert not is_namedtuple(Tuple[int])

    class Test(NamedTuple):
        a: int

    assert is_namedtuple(Test)

    class Test2(Tuple[int, int]):
        pass

    assert not is_namedtuple(Test2)


def test_simple_typed_namedtuples(genconverter: Converter):
    """Simple typed namedtuples work."""

    class Test(NamedTuple):
        a: int

    assert genconverter.unstructure(Test(1)) == Test(1)
    assert genconverter.structure([1], Test) == Test(1)

    genconverter.register_unstructure_hook(int, lambda v: v + 1)
    genconverter.register_structure_hook(int, lambda v, _: v - 1)

    assert genconverter.unstructure(Test(1)) == (2,)
    assert genconverter.structure([2], Test) == Test(1)


def test_simple_dict_nametuples(genconverter: Converter):
    """Namedtuples can be un/structured to/from dicts."""

    class Test(NamedTuple):
        a: int
        b: str = "test"

    genconverter.register_unstructure_hook_factory(
        lambda t: t is Test, namedtuple_dict_unstructure_factory
    )
    genconverter.register_structure_hook_factory(
        lambda t: t is Test, namedtuple_dict_structure_factory
    )

    assert genconverter.unstructure(Test(1)) == {"a": 1, "b": "test"}
    assert genconverter.structure({"a": 1, "b": "2"}, Test) == Test(1, "2")

    # Defaults work.
    assert genconverter.structure({"a": 1}, Test) == Test(1, "test")


@define
class RecursiveAttrs:
    b: "List[RecursiveNamedtuple]" = Factory(list)


class RecursiveNamedtuple(NamedTuple):
    a: RecursiveAttrs


def test_recursive_dict_nametuples(genconverter: Converter):
    """Recursive namedtuples can be un/structured to/from dicts."""

    genconverter.register_unstructure_hook_factory(
        lambda t: t is RecursiveNamedtuple, namedtuple_dict_unstructure_factory
    )
    genconverter.register_structure_hook_factory(
        lambda t: t is RecursiveNamedtuple, namedtuple_dict_structure_factory
    )

    assert genconverter.unstructure(RecursiveNamedtuple(RecursiveAttrs())) == {
        "a": {"b": []}
    }
    assert genconverter.structure(
        {"a": {}}, RecursiveNamedtuple
    ) == RecursiveNamedtuple(RecursiveAttrs())


def test_dict_nametuples_forbid_extra_keys(genconverter: Converter):
    """Forbidding extra keys works for structuring namedtuples from dicts."""

    class Test(NamedTuple):
        a: int

    genconverter.register_structure_hook_factory(
        lambda t: t is Test,
        lambda t, c: namedtuple_dict_structure_factory(t, c, "from_converter", True),
    )

    with raises(Exception) as exc_info:
        genconverter.structure({"a": 1, "b": "2"}, Test)

    if genconverter.detailed_validation:
        exc = exc_info.value.exceptions[0]
    else:
        exc = exc_info.value

    assert isinstance(exc, ForbiddenExtraKeysError)
    assert exc.extra_fields == {"b"}



# cattrs documentation build configuration file, created by
# sphinx-quickstart on Tue Jul  9 22:26:36 2013.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys
from importlib.metadata import version as v

# Set canonical URL from the Read the Docs Domain
html_baseurl = os.environ.get("READTHEDOCS_CANONICAL_URL", "")

# Tell Jinja2 templates the build is running on Read the Docs
if os.environ.get("READTHEDOCS", "") == "True":
    html_context = {"READTHEDOCS": True}


# If extensions (or modules to document with autodoc) are in another
# directory, add these directories to sys.path here. If the directory is
# relative to the documentation root, use os.path.abspath to make it
# absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

# Get the project root dir, which is the parent dir of this
cwd = os.getcwd()
project_root = os.path.join(os.path.dirname(cwd), "src")

# Insert the project root dir as the first element in the PYTHONPATH.
# This lets us ensure that the source package is imported, and that its
# version is used.
sys.path.insert(0, project_root)

# -- General configuration ---------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.doctest",
    "sphinx_copybutton",
    "myst_parser",
]

myst_enable_extensions = ["colon_fence", "smartquotes", "deflist"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "cattrs"
copyright = "2020, Tin Tvrtković"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = v("cattrs")
# The full version, including alpha/beta/rc tags.
release = version
if "dev" in release:
    release = version = "UNRELEASED"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to
# some non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
# pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []

# If true, keep warnings as "system message" paragraphs in the built
# documents.
# keep_warnings = False


# -- Options for HTML output -------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "furo"
html_theme_options = {
    "light_css_variables": {
        "font-stack": "Inter,sans-serif",
        "font-stack--monospace": "'Ubuntu Mono', monospace",
        "code-font-size": "90%",
        "color-highlight-on-target": "transparent",
    },
    "dark_css_variables": {"color-highlight-on-target": "transparent"},
}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as
# html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the
# top of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon
# of the docs.  This file should be a Windows icon file (.ico) being
# 16x16 or 32x32 pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets)
# here, relative to this directory. They are copied after the builtin
# static files, so a file named "default.css" will overwrite the builtin
# "default.css".
html_static_path = ["_static"]

html_css_files = ["custom.css"]

# If not '', a 'Last updated on:' timestamp is inserted at every page
# bottom, using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names
# to template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer.
# Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer.
# Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages
# will contain a <link> tag referring to it.  The value of this option
# must be the base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "cattrsdoc"


# -- Options for LaTeX output ------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',
    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',
    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    ("index", "cattrs.tex", "cattrs Documentation", "Tin Tvrtković", "manual")
]

# The name of an image file (relative to this directory) to place at
# the top of the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings
# are parts, not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [("index", "cattrs", "cattrs Documentation", ["Tin Tvrtković"], 1)]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "cattrs",
        "cattrs Documentation",
        "Tin Tvrtković",
        "cattrs",
        "Composable complex class support for attrs.",
        "Miscellaneous",
    )
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'

# If true, do not generate a @detailmenu in the "Top" node's menu.
# texinfo_no_detailmenu = False

doctest_global_setup = (
    "import attr, cattr, cattrs;"
    "from attr import Factory, define, field;"
    "from cattrs import Converter;"
    "from typing import *;"
    "from enum import Enum, unique"
)
autodoc_member_order = "bysource"
autodoc_typehints = "description"
autosectionlabel_prefix_document = True
copybutton_prompt_text = r">>> |\.\.\. "
copybutton_prompt_is_regexp = True
myst_heading_anchors = 3
autoclass_content = "both"



"""Tests for generics under PEP 604 (unions as pipes)."""

from typing import Generic, TypeVar

from attrs import define

T = TypeVar("T")


def test_unstructure_optional(genconverter):
    """Generics with optional fields work."""

    @define
    class C(Generic[T]):
        a: T | None

    assert genconverter.unstructure(C(C(1))) == {"a": {"a": 1}}



"""Tests for auto-disambiguators."""

from dataclasses import dataclass
from functools import partial
from typing import Literal, Union

import pytest
from attrs import NOTHING, asdict, define, field, fields
from hypothesis import HealthCheck, assume, given, settings

from cattrs import Converter
from cattrs.disambiguators import create_default_dis_func, is_supported_union
from cattrs.gen import make_dict_structure_fn, override

from .untyped import simple_classes


def test_edge_errors():
    """Edge input cases cause errors."""
    c = Converter()

    @define
    class A:
        pass

    with pytest.raises(ValueError):
        # Can't generate for only one class.
        create_default_dis_func(c, A)

    with pytest.raises(ValueError):
        create_default_dis_func(c, A)

    @define
    class B:
        pass

    with pytest.raises(TypeError):
        # No fields on either class.
        create_default_dis_func(c, A, B)

    @define
    class C:
        a = field()

    @define
    class D:
        a = field()

    with pytest.raises(TypeError):
        # No unique fields on either class.
        create_default_dis_func(c, C, D)

    with pytest.raises(TypeError):
        # No discriminator candidates
        create_default_dis_func(c, C, D)

    @define
    class E:
        pass

    @define
    class F:
        b = None

    with pytest.raises(TypeError):
        # no usable non-default attributes
        create_default_dis_func(c, E, F)

    @define
    class G:
        x: Literal[1]

    @define
    class H:
        x: Literal[1]

    with pytest.raises(TypeError):
        # The discriminator chosen does not actually help
        create_default_dis_func(c, C, D)


@given(simple_classes(defaults=False))
def test_fallback(cl_and_vals):
    """The fallback case works."""
    cl, vals, kwargs = cl_and_vals
    c = Converter()

    assume(fields(cl))  # At least one field.

    @define
    class A:
        pass

    fn = create_default_dis_func(c, A, cl)

    assert fn({}) is A
    assert fn(asdict(cl(*vals, **kwargs))) is cl

    attr_names = {a.name for a in fields(cl)}

    if "xyz" not in attr_names:
        assert fn({"xyz": 1}) is A  # Uses the fallback.


@settings(suppress_health_check=[HealthCheck.filter_too_much, HealthCheck.too_slow])
@given(simple_classes(), simple_classes())
def test_disambiguation(cl_and_vals_a, cl_and_vals_b):
    """Disambiguation should work when there are unique required fields."""
    cl_a, vals_a, kwargs_a = cl_and_vals_a
    cl_b, vals_b, kwargs_b = cl_and_vals_b
    c = Converter()

    req_a = {a.name for a in fields(cl_a)}
    req_b = {a.name for a in fields(cl_b)}

    assume(len(req_a))
    assume(len(req_b))

    assume((req_a - req_b) or (req_b - req_a))
    for attr_name in req_a - req_b:
        assume(getattr(fields(cl_a), attr_name).default is NOTHING)
    for attr_name in req_b - req_a:
        assume(getattr(fields(cl_b), attr_name).default is NOTHING)

    fn = create_default_dis_func(c, cl_a, cl_b)

    assert fn(asdict(cl_a(*vals_a, **kwargs_a))) is cl_a


# not too sure of properties of `create_default_dis_func`
def test_disambiguate_from_discriminated_enum():
    c = Converter()

    # can it find any discriminator?
    @define
    class A:
        a: Literal[0]

    @define
    class B:
        a: Literal[1]

    fn = create_default_dis_func(c, A, B)
    assert fn({"a": 0}) is A
    assert fn({"a": 1}) is B

    # can it find the better discriminator?
    @define
    class C:
        a: Literal[0]
        b: Literal[1]

    @define
    class D:
        a: Literal[0]
        b: Literal[0]

    fn = create_default_dis_func(c, C, D)
    assert fn({"a": 0, "b": 1}) is C
    assert fn({"a": 0, "b": 0}) is D

    # can it handle multiple tiers of discriminators?
    # (example inspired by Discord's gateway's discriminated union)
    @define
    class E:
        op: Literal[1]

    @define
    class F:
        op: Literal[0]
        t: Literal["MESSAGE_CREATE"]

    @define
    class G:
        op: Literal[0]
        t: Literal["MESSAGE_UPDATE"]

    fn = create_default_dis_func(c, E, F, G)
    assert fn({"op": 1}) is E
    assert fn({"op": 0, "t": "MESSAGE_CREATE"}) is Union[F, G]

    # can it handle multiple literals?
    @define
    class H:
        a: Literal[1]

    @define
    class J:
        a: Literal[0, 1]

    @define
    class K:
        a: Literal[0]

    fn = create_default_dis_func(c, H, J, K)
    assert fn({"a": 1}) is Union[H, J]
    assert fn({"a": 0}) is Union[J, K]


def test_default_no_literals():
    """The default disambiguator can skip literals."""
    c = Converter()

    @define
    class A:
        a: Literal["a"] = "a"

    @define
    class B:
        a: Literal["b"] = "b"

    default = create_default_dis_func(c, A, B)  # Should work.
    assert default({"a": "a"}) is A

    with pytest.raises(TypeError):
        create_default_dis_func(c, A, B, use_literals=False)

    @define
    class C:
        b: int
        a: Literal["a"] = "a"

    @define
    class D:
        a: Literal["b"] = "b"

    default = create_default_dis_func(c, C, D)  # Should work.
    assert default({"a": "a"}) is C

    no_lits = create_default_dis_func(c, C, D, use_literals=False)
    assert no_lits({"a": "a", "b": 1}) is C
    assert no_lits({"a": "a"}) is D


def test_converter_no_literals(converter: Converter):
    """A converter can be configured to skip literals."""

    converter.register_structure_hook_factory(
        is_supported_union,
        partial(converter._gen_attrs_union_structure, use_literals=False),
    )

    @define
    class C:
        b: int
        a: Literal["a"] = "a"

    @define
    class D:
        a: Literal["b"] = "b"

    assert converter.structure({}, Union[C, D]) == D()


def test_field_renaming(converter: Converter):
    """A renamed field properly disambiguates."""

    @define
    class A:
        a: int

    @define
    class B:
        a: int

    converter.register_structure_hook(
        B, make_dict_structure_fn(B, converter, a=override(rename="b"))
    )

    assert converter.structure({"a": 1}, Union[A, B]) == A(1)
    assert converter.structure({"b": 1}, Union[A, B]) == B(1)


def test_dataclasses(converter):
    """The default strategy works for dataclasses too."""

    @define
    class A:
        a: int

    @dataclass
    class B:
        b: int

    assert converter.structure({"a": 1}, Union[A, B]) == A(1)
    assert converter.structure({"b": 1}, Union[A, B]) == B(1)


def test_dataclasses_literals(converter):
    """The default strategy works for dataclasses too."""

    @define
    class A:
        a: Literal["a"] = "a"

    @dataclass
    class B:
        b: Literal["b"]

    assert converter.structure({"a": "a"}, Union[A, B]) == A()
    assert converter.structure({"b": "b"}, Union[A, B]) == B("b")



"""Tests for enums."""

from hypothesis import given
from hypothesis.strategies import data, sampled_from
from pytest import raises

from cattrs import BaseConverter
from cattrs._compat import Literal

from .untyped import enums_of_primitives


@given(data(), enums_of_primitives())
def test_structuring_enums(data, enum):
    """Test structuring enums by their values."""
    converter = BaseConverter()
    val = data.draw(sampled_from(list(enum)))

    assert converter.structure(val.value, enum) == val


@given(enums_of_primitives())
def test_enum_failure(enum):
    """Structuring literals with enums fails properly."""
    converter = BaseConverter()
    type = Literal[next(iter(enum))]

    with raises(Exception) as exc_info:
        converter.structure("", type)

    assert exc_info.value.args[0] == f" not in literal {type!r}"



"""Tests for TypedDict un/structuring."""

from datetime import datetime, timezone
from typing import Dict, Generic, NewType, Set, Tuple, TypedDict, TypeVar

import pytest
from attrs import NOTHING
from hypothesis import assume, given
from hypothesis.strategies import booleans
from pytest import raises
from typing_extensions import NotRequired, Required

from cattrs import BaseConverter, Converter
from cattrs._compat import ExtensionsTypedDict, get_notrequired_base, is_generic
from cattrs.errors import (
    ClassValidationError,
    ForbiddenExtraKeysError,
    StructureHandlerNotFoundError,
)
from cattrs.gen import already_generating, override
from cattrs.gen._generics import generate_mapping
from cattrs.gen.typeddicts import (
    get_annots,
    make_dict_structure_fn,
    make_dict_unstructure_fn,
)

from ._compat import is_py38, is_py39, is_py310, is_py311_plus
from .typeddicts import (
    generic_typeddicts,
    simple_typeddicts,
    simple_typeddicts_with_extra_keys,
)


def mk_converter(detailed_validation: bool = True) -> Converter:
    """We can't use function-scoped fixtures with Hypothesis strats."""
    c = Converter(detailed_validation=detailed_validation)
    c.register_unstructure_hook(datetime, lambda d: d.timestamp())
    c.register_structure_hook(
        datetime, lambda d, _: datetime.fromtimestamp(d, tz=timezone.utc)
    )
    return c


def get_annot(t) -> dict:
    """Our version, handling type vars properly."""
    if is_generic(t):
        # This will have typevars.
        origin = getattr(t, "__origin__", None)
        if origin is not None:
            origin_annotations = get_annots(origin)
            args = t.__args__
            params = origin.__parameters__
            param_to_args = dict(zip(params, args))
            res = {}
            for k, v in origin_annotations.items():
                if (nrb := get_notrequired_base(v)) is not NOTHING:
                    res[k] = (
                        NotRequired[param_to_args[nrb]] if nrb in param_to_args else v
                    )
                else:
                    res[k] = param_to_args[v] if v in param_to_args else v
            return res

        # Origin is `None`, so this is a subclass for a generic typeddict.
        mapping = generate_mapping(t)
        res = {}
        for k, v in get_annots(t).items():
            if (nrb := get_notrequired_base(v)) is not NOTHING:
                res[k] = (
                    NotRequired[mapping[nrb.__name__]] if nrb.__name__ in mapping else v
                )
            else:
                res[k] = mapping[v.__name__] if v.__name__ in mapping else v
        return res
    return get_annots(t)


@given(simple_typeddicts(typeddict_cls=None if not is_py38 else ExtensionsTypedDict))
def test_simple_roundtrip(cls_and_instance) -> None:
    """Round-trips for simple classes work."""
    c = mk_converter()
    cls, instance = cls_and_instance

    unstructured = c.unstructure(instance, unstructure_as=cls)

    if all(a is not datetime for _, a in get_annot(cls).items()):
        assert unstructured == instance

    if all(a is int for _, a in get_annot(cls).items()):
        assert unstructured is instance

    restructured = c.structure(unstructured, cls)

    assert restructured is not unstructured
    assert restructured == instance


@given(
    simple_typeddicts(
        total=False, typeddict_cls=None if not is_py38 else ExtensionsTypedDict
    ),
    booleans(),
)
def test_simple_nontotal(cls_and_instance, detailed_validation: bool) -> None:
    """Non-total dicts work."""
    c = mk_converter(detailed_validation=detailed_validation)
    cls, instance = cls_and_instance

    unstructured = c.unstructure(instance, unstructure_as=cls)

    if all(a is not datetime for _, a in get_annot(cls).items()):
        assert unstructured == instance

    if all(a is int for _, a in get_annot(cls).items()):
        assert unstructured is instance

    restructured = c.structure(unstructured, cls)

    assert restructured is not unstructured
    assert restructured == instance


@given(simple_typeddicts(typeddict_cls=None if not is_py38 else ExtensionsTypedDict))
def test_int_override(cls_and_instance) -> None:
    """Overriding a base unstructure handler should work."""
    cls, instance = cls_and_instance

    assume(any(a is int for _, a in get_annot(cls).items()))
    assume(all(a is not datetime for _, a in get_annot(cls).items()))

    c = mk_converter()
    c.register_unstructure_hook(int, lambda i: i)
    unstructured = c.unstructure(instance, unstructure_as=cls)

    assert unstructured is not instance
    assert unstructured == instance


@given(
    simple_typeddicts_with_extra_keys(
        typeddict_cls=None if not is_py38 else ExtensionsTypedDict
    ),
    booleans(),
)
def test_extra_keys(
    cls_instance_extra: Tuple[type, Dict, Set[str]], detailed_validation: bool
) -> None:
    """Extra keys are preserved."""
    cls, instance, extra = cls_instance_extra

    c = mk_converter(detailed_validation)

    unstructured = c.unstructure(instance, unstructure_as=cls)
    for k in extra:
        assert k in unstructured

    structured = c.structure(unstructured, cls)

    for k in extra:
        assert k in structured

    assert structured == instance


@pytest.mark.skipif(not is_py311_plus, reason="3.11+ only")
@given(generic_typeddicts(total=True), booleans())
def test_generics(
    cls_and_instance: Tuple[type, Dict], detailed_validation: bool
) -> None:
    """Generic TypedDicts work."""
    c = mk_converter(detailed_validation=detailed_validation)
    cls, instance = cls_and_instance

    unstructured = c.unstructure(instance, unstructure_as=cls)
    assert not any(isinstance(v, datetime) for v in unstructured.values())

    if all(
        a not in (datetime, NotRequired[datetime]) for _, a in get_annot(cls).items()
    ):
        assert unstructured == instance

    if all(a is int for _, a in get_annot(cls).items()):
        assert unstructured is instance

    restructured = c.structure(unstructured, cls)

    assert restructured is not unstructured
    assert restructured == instance


@pytest.mark.skipif(not is_py311_plus, reason="3.11+ only")
@given(booleans())
def test_generics_with_unbound(detailed_validation: bool):
    """TypedDicts with unbound TypeVars work."""
    c = mk_converter(detailed_validation=detailed_validation)

    T = TypeVar("T")

    class GenericTypedDict(TypedDict, Generic[T]):
        a: T

    assert c.unstructure({"a": 1}, GenericTypedDict)

    with pytest.raises(StructureHandlerNotFoundError):
        # This doesn't work since we refuse the temptation to guess.
        c.structure({"a": 1}, GenericTypedDict)


@pytest.mark.skipif(not is_py311_plus, reason="3.11+ only")
@given(detailed_validation=...)
def test_deep_generics(detailed_validation: bool):
    c = mk_converter(detailed_validation=detailed_validation)

    Int = NewType("Int", int)

    c.register_unstructure_hook_func(lambda t: t is Int, lambda v: v - 1)

    T = TypeVar("T")
    T1 = TypeVar("T1")

    class GenericParent(TypedDict, Generic[T]):
        a: T

    class GenericChild(GenericParent[Int], Generic[T1]):
        b: T1

    assert c.unstructure({"b": 2, "a": 2}, GenericChild[Int]) == {"a": 1, "b": 1}


@given(simple_typeddicts(total=True, not_required=True), booleans())
def test_not_required(
    cls_and_instance: Tuple[type, Dict], detailed_validation: bool
) -> None:
    """NotRequired[] keys are handled."""
    c = mk_converter(detailed_validation=detailed_validation)
    cls, instance = cls_and_instance

    unstructured = c.unstructure(instance, unstructure_as=cls)
    restructured = c.structure(unstructured, cls)

    assert restructured == instance


@given(
    simple_typeddicts(
        total=False,
        not_required=True,
        typeddict_cls=None if not is_py38 else ExtensionsTypedDict,
    ),
    booleans(),
)
def test_required(
    cls_and_instance: Tuple[type, Dict], detailed_validation: bool
) -> None:
    """Required[] keys are handled."""
    c = mk_converter(detailed_validation=detailed_validation)
    cls, instance = cls_and_instance

    unstructured = c.unstructure(instance, unstructure_as=cls)
    restructured = c.structure(unstructured, cls)

    assert restructured == instance


@pytest.mark.skipif(is_py39 or is_py310, reason="Sigh")
def test_required_keys() -> None:
    """We don't support the full gamut of functionality on 3.8.

    When using `typing.TypedDict` we have only partial functionality;
    this test tests only a subset of this.
    """
    c = mk_converter()

    class Base(TypedDict, total=False):
        a: Required[datetime]

    class Sub(Base):
        b: int

    fn = make_dict_unstructure_fn(Sub, c)

    with raises(KeyError):
        # This needs to raise since 'a' is missing, and it's Required.
        fn({"b": 1})


@given(simple_typeddicts(min_attrs=1, total=True), booleans())
def test_omit(cls_and_instance: Tuple[type, Dict], detailed_validation: bool) -> None:
    """`override(omit=True)` works."""
    c = mk_converter(detailed_validation=detailed_validation)

    cls, instance = cls_and_instance
    key = next(iter(get_annot(cls)))
    c.register_unstructure_hook(
        cls,
        make_dict_unstructure_fn(
            cls,
            c,
            _cattrs_detailed_validation=detailed_validation,
            **{key: override(omit=True)},
        ),
    )

    unstructured = c.unstructure(instance, unstructure_as=cls)

    assert key not in unstructured

    unstructured[key] = c.unstructure(instance[key])
    restructured = c.structure(unstructured, cls)

    assert restructured == instance

    c.register_structure_hook(
        cls,
        make_dict_structure_fn(
            cls,
            c,
            _cattrs_detailed_validation=detailed_validation,
            **{key: override(omit=True)},
        ),
    )
    del unstructured[key]
    del instance[key]
    restructured = c.structure(unstructured, cls)

    assert restructured == instance


@given(simple_typeddicts(min_attrs=1, total=True, not_required=True), booleans())
def test_rename(cls_and_instance: Tuple[type, Dict], detailed_validation: bool) -> None:
    """`override(rename=...)` works."""
    c = mk_converter(detailed_validation=detailed_validation)

    cls, instance = cls_and_instance
    key = next(iter(get_annot(cls)))
    c.register_unstructure_hook(
        cls, make_dict_unstructure_fn(cls, c, **{key: override(rename="renamed")})
    )

    unstructured = c.unstructure(instance, unstructure_as=cls)

    assert key not in unstructured
    if key in instance:
        assert "renamed" in unstructured

    c.register_structure_hook(
        cls, make_dict_structure_fn(cls, c, **{key: override(rename="renamed")})
    )
    restructured = c.structure(unstructured, cls)

    assert restructured == instance


@given(simple_typeddicts(total=True), booleans())
def test_forbid_extra_keys(
    cls_and_instance: Tuple[type, Dict], detailed_validation: bool
) -> None:
    """Extra keys can be forbidden."""
    c = mk_converter(detailed_validation)

    cls, instance = cls_and_instance

    c.register_structure_hook(
        cls,
        make_dict_structure_fn(
            cls,
            c,
            _cattrs_detailed_validation=detailed_validation,
            _cattrs_forbid_extra_keys=True,
        ),
    )

    unstructured = c.unstructure(instance, unstructure_as=cls)

    structured = c.structure(unstructured, cls)
    assert structured == instance

    # An extra key will trigger the appropriate error.
    unstructured["test"] = 1

    if not detailed_validation:
        with raises(ForbiddenExtraKeysError):
            c.structure(unstructured, cls)
    else:
        with raises(ClassValidationError) as ctx:
            c.structure(unstructured, cls)

        assert repr(ctx.value) == repr(
            ClassValidationError(
                f"While structuring {cls.__name__}",
                [
                    ForbiddenExtraKeysError(
                        f"Extra fields in constructor for {cls.__name__}: test",
                        cls,
                        {"test"},
                    )
                ],
                cls,
            )
        )


class TypedDictA(ExtensionsTypedDict):
    b: "TypedDictB"


class TypedDictB(ExtensionsTypedDict):
    a: "TypedDictA"


@given(...)
def test_recursive_generation(detailed_validation: bool) -> None:
    """Generating recursive hooks works."""
    assert not hasattr(already_generating, "working_set")

    c = mk_converter(detailed_validation)

    assert c.unstructure({"a": {"b": {"a": {}}}}, TypedDictB) == {"a": {"b": {"a": {}}}}


def test_forwardref(genconverter: Converter):
    """TypedDicts have no resolve_class, so they're good candidate for forwardrefs."""

    class A(ExtensionsTypedDict):
        a: "int"

    genconverter.register_unstructure_hook(int, lambda v: v + 1)

    assert genconverter.unstructure({"a": 1}, A) == {"a": 2}


@given(forbid_extra_keys=..., detailed_validation=...)
def test_forbid_extra_keys_from_converter(
    forbid_extra_keys: bool, detailed_validation: bool
):
    """
    `forbid_extra_keys` is taken from the converter by default.
    """
    c = Converter(
        forbid_extra_keys=forbid_extra_keys, detailed_validation=detailed_validation
    )

    class A(ExtensionsTypedDict):
        a: int

    c.register_structure_hook(A, make_dict_structure_fn(A, c))

    if forbid_extra_keys:
        with pytest.raises((ForbiddenExtraKeysError, ClassValidationError)):
            c.structure({"a": 1, "b": 2}, A)
    else:
        c.structure({"a": 1, "b": 2}, A)


@given(detailed_validation=...)
def test_forbid_extra_keys_from_baseconverter(detailed_validation: bool):
    """
    `forbid_extra_keys` is taken from the converter by default.

    BaseConverter should default to False.
    """
    c = BaseConverter(detailed_validation=detailed_validation)

    class A(ExtensionsTypedDict):
        a: int

    c.register_structure_hook(A, make_dict_structure_fn(A, c))

    c.structure({"a": 1, "b": 2}, A)


def test_detailed_validation_from_converter(converter: BaseConverter):
    """
    `detailed_validation` is taken from the converter by default.
    """

    class A(ExtensionsTypedDict):
        a: int

    converter.register_structure_hook(A, make_dict_structure_fn(A, converter))

    if converter.detailed_validation:
        with pytest.raises(ClassValidationError):
            converter.structure({"a": "a"}, A)
    else:
        with pytest.raises(ValueError):
            converter.structure({"a": "a"}, A)


def test_override_entire_hooks(converter: BaseConverter):
    """Overriding entire hooks works."""

    class A(ExtensionsTypedDict):
        a: int
        b: NotRequired[int]

    converter.register_structure_hook(
        A,
        make_dict_structure_fn(
            A,
            converter,
            a=override(struct_hook=lambda v, _: 1),
            b=override(struct_hook=lambda v, _: 2),
        ),
    )
    converter.register_unstructure_hook(
        A,
        make_dict_unstructure_fn(
            A,
            converter,
            a=override(unstruct_hook=lambda v: 1),
            b=override(unstruct_hook=lambda v: 2),
        ),
    )

    assert converter.unstructure({"a": 10, "b": 10}, A) == {"a": 1, "b": 2}
    assert converter.structure({"a": 10, "b": 10}, A) == {"a": 1, "b": 2}



"""Pyyaml-specific tests."""

from datetime import date, datetime, timezone

from attrs import define
from hypothesis import given
from pytest import raises

from cattrs._compat import FrozenSetSubscriptable
from cattrs.errors import ClassValidationError
from cattrs.preconf.pyyaml import make_converter

from .._compat import is_py38
from ..test_preconf import Everything, everythings, native_unions


@given(everythings())
def test_pyyaml(everything: Everything):
    from yaml import safe_dump, safe_load

    converter = make_converter()
    unstructured = converter.unstructure(everything)
    raw = safe_dump(unstructured)
    assert converter.structure(safe_load(raw), Everything) == everything


@given(everythings())
def test_pyyaml_converter(everything: Everything):
    converter = make_converter()
    raw = converter.dumps(everything)
    assert converter.loads(raw, Everything) == everything


@given(everythings())
def test_pyyaml_converter_unstruct_collection_overrides(everything: Everything):
    converter = make_converter(
        unstruct_collection_overrides={FrozenSetSubscriptable: sorted}
    )
    raw = converter.unstructure(everything)
    assert raw["a_frozenset"] == sorted(raw["a_frozenset"])


@given(
    union_and_val=native_unions(include_bools=not is_py38),  # Literal issues on 3.8
    detailed_validation=...,
)
def test_pyyaml_unions(union_and_val: tuple, detailed_validation: bool):
    """Native union passthrough works."""
    converter = make_converter(detailed_validation=detailed_validation)
    type, val = union_and_val

    assert converter.structure(val, type) == val


@given(detailed_validation=...)
def test_pyyaml_dates(detailed_validation: bool):
    """Pyyaml dates work."""
    converter = make_converter(detailed_validation=detailed_validation)

    @define
    class A:
        datetime: datetime
        date: date

    data = """
    datetime: 1970-01-01T00:00:00Z
    date: 1970-01-01"""
    assert converter.loads(data, A) == A(
        datetime(1970, 1, 1, tzinfo=timezone.utc), date(1970, 1, 1)
    )

    bad_data = """
    datetime: 1
    date: 1
    """

    with raises(ClassValidationError if detailed_validation else Exception) as exc_info:
        converter.loads(bad_data, A)

    if detailed_validation:
        assert (
            repr(exc_info.value.exceptions[0])
            == "Exception('Expected datetime, got 1')"
        )
        assert (
            repr(exc_info.value.exceptions[1]) == "ValueError('Expected date, got 1')"
        )



from typing import Any, Dict


def test_any_keys(converter):
    """Dicts with any keys work."""
    assert converter.structure({b"": "1"}, Dict[Any, int]) == {b"": 1}


def test_any_values(converter):
    """Dicts with any values work."""
    assert converter.structure({"1": b"1"}, Dict[int, Any]) == {1: b"1"}



"""Tests for the native union passthrough strategy.

Note that a significant amount of test coverage for this is in the
preconf tests.
"""

from typing import List, Optional, Union

import pytest
from attrs import define

from cattrs import BaseConverter
from cattrs.strategies import configure_union_passthrough


def test_only_primitives(converter: BaseConverter) -> None:
    """A native union with only primitives works."""
    union = Union[int, str, None]
    configure_union_passthrough(union, converter)

    assert converter.unstructure(1, union) == 1
    assert converter.structure(1, union) == 1
    assert converter.unstructure("1", union) == "1"
    assert converter.structure("1", union) == "1"
    assert converter.unstructure(None, union) is None
    assert converter.structure(None, union) is None

    with pytest.raises(TypeError):
        converter.structure((), union)


def test_literals(converter: BaseConverter) -> None:
    """A union with primitives and literals works."""
    from typing import Literal

    union = Union[int, str, None]
    exact_type = Union[int, Literal["test"], None]
    configure_union_passthrough(union, converter)

    assert converter.unstructure(1, exact_type) == 1
    assert converter.structure(1, exact_type) == 1
    assert converter.unstructure("test", exact_type) == "test"
    assert converter.structure("test", exact_type) == "test"
    assert converter.unstructure(None, exact_type) is None
    assert converter.structure(None, exact_type) is None

    with pytest.raises(TypeError):
        converter.structure((), exact_type)
    with pytest.raises(TypeError):
        converter.structure("t", exact_type)


def test_skip_optionals() -> None:
    """
    The strategy skips Optionals, since those are more efficiently
    handled by default.
    """
    c = BaseConverter()

    configure_union_passthrough(Union[int, str, None], c)

    h = c.get_structure_hook(Optional[int])
    assert h.__name__ != "structure_native_union"


def test_spillover(converter: BaseConverter) -> None:
    """Types not covered by the native union are correctly handled."""
    union = Union[int, str, None]
    exact_type = Union[int, List[str], None]

    configure_union_passthrough(union, converter)

    assert converter.unstructure(1, exact_type) == 1
    assert converter.structure(1, exact_type) == 1

    assert converter.unstructure(["a", "b"], exact_type) == ["a", "b"]
    assert converter.structure(["a", "b"], exact_type) == ["a", "b"]

    with pytest.raises(TypeError):
        converter.structure((), union)


def test_multiple_spillover(converter: BaseConverter) -> None:
    """Types not covered by the native union are correctly handled."""
    union = Union[int, str, None]

    @define
    class A:
        a: int

    @define
    class B:
        b: int

    # A | B will be handled by the default disambiguator.
    exact_type = Union[int, List[str], A, B, None]

    configure_union_passthrough(union, converter)

    assert converter.unstructure(1, exact_type) == 1
    assert converter.structure(1, exact_type) == 1

    assert converter.unstructure(["a", "b"], exact_type) == ["a", "b"]
    assert converter.structure(["a", "b"], List[str]) == ["a", "b"]
    assert converter.unstructure(A(1), exact_type) == {"a": 1}
    assert converter.structure({"a": 1}, A) == A(1)
    assert converter.unstructure(B(1), exact_type) == {"b": 1}
    assert converter.structure({"b": 1}, B) == B(1)

    with pytest.raises(TypeError):
        converter.structure((), union)



"""Tests for generated dict functions."""

from typing import Dict, Type

import pytest
from attrs import NOTHING, Factory, define, field, frozen
from hypothesis import assume, given
from hypothesis.strategies import data, just, one_of, sampled_from

from cattrs import BaseConverter, Converter
from cattrs._compat import adapted_fields, fields
from cattrs.errors import ClassValidationError, ForbiddenExtraKeysError
from cattrs.gen import make_dict_structure_fn, make_dict_unstructure_fn, override

from ._compat import is_py39_plus
from .typed import nested_typed_classes, simple_typed_classes, simple_typed_dataclasses
from .untyped import nested_classes, simple_classes


@given(nested_classes | simple_classes())
def test_unmodified_generated_unstructuring(cl_and_vals):
    converter = BaseConverter()
    cl, vals, kwargs = cl_and_vals
    fn = make_dict_unstructure_fn(cl, converter)

    inst = cl(*vals, **kwargs)

    res_expected = converter.unstructure(inst)

    converter.register_unstructure_hook(cl, fn)

    res_actual = converter.unstructure(inst)

    assert res_expected == res_actual


@given(nested_classes | simple_classes())
def test_nodefs_generated_unstructuring(cl_and_vals):
    """Test omitting default values on a per-attribute basis."""
    converter = BaseConverter()
    cl, vals, kwargs = cl_and_vals

    attr_is_default = False
    for attr, val in zip([a for a in cl.__attrs_attrs__ if not a.kw_only], vals):
        if attr.default is not NOTHING:
            fn = make_dict_unstructure_fn(
                cl, converter, **{attr.name: override(omit_if_default=True)}
            )
            if attr.default == val:
                attr_is_default = True
            break
    else:
        assume(False)

    converter.register_unstructure_hook(cl, fn)

    inst = cl(*vals, **kwargs)

    res = converter.unstructure(inst)

    if attr_is_default:
        assert attr.name not in res


@given(one_of(just(BaseConverter), just(Converter)), nested_classes | simple_classes())
def test_nodefs_generated_unstructuring_cl(
    converter_cls: Type[BaseConverter], cl_and_vals
):
    """Test omitting default values on a per-class basis."""
    converter = converter_cls()
    cl, vals, kwargs = cl_and_vals

    for attr in cl.__attrs_attrs__:
        if attr.default is not NOTHING:
            break
    else:
        assume(False)

    converter.register_unstructure_hook(
        cl, make_dict_unstructure_fn(cl, converter, _cattrs_omit_if_default=True)
    )

    inst = cl(*vals, **kwargs)

    res = converter.unstructure(inst)

    for attr, val in zip(cl.__attrs_attrs__, vals):
        if attr.default is not NOTHING:
            if not isinstance(attr.default, Factory):
                if val == attr.default:
                    assert attr.name not in res
                else:
                    assert attr.name in res
            else:
                # The default is a factory, but might take self.
                if attr.default.takes_self:
                    # Our strategies can only produce these for now.
                    assert val == attr.default.factory(cl)
                    assert attr.name not in res
                else:
                    if val == attr.default.factory():
                        assert attr.name not in res
                    else:
                        assert attr.name in res


@given(
    one_of(just(BaseConverter), just(Converter)),
    nested_classes | simple_classes() | simple_typed_dataclasses(),
)
def test_individual_overrides(converter_cls, cl_and_vals):
    """
    Test omitting default values on a per-class basis, but with individual
    overrides.
    """
    converter = converter_cls()
    cl, vals, kwargs = cl_and_vals

    for attr in adapted_fields(cl):
        if attr.default is not NOTHING:
            break
    else:
        assume(False)

    chosen_name = attr.name

    converter.register_unstructure_hook(
        cl,
        make_dict_unstructure_fn(
            cl,
            converter,
            _cattrs_omit_if_default=True,
            **{attr.name: override(omit_if_default=False)},
        ),
    )

    inst = cl(*vals, **kwargs)

    res = converter.unstructure(inst)
    assert "Hyp" not in repr(res)
    assert "Factory" not in repr(res)

    for attr, val in zip(adapted_fields(cl), vals):
        if attr.name == chosen_name:
            assert attr.name in res
        elif attr.default is not NOTHING:
            if not isinstance(attr.default, Factory):
                if val == attr.default:
                    assert attr.name not in res
                else:
                    assert attr.name in res
            else:
                if attr.default.takes_self:
                    assert attr.name not in res
                else:
                    if val == attr.default.factory():
                        assert attr.name not in res
                    else:
                        assert attr.name in res


@given(
    cl_and_vals=nested_typed_classes()
    | simple_typed_classes()
    | simple_typed_dataclasses(),
    dv=...,
)
def test_unmodified_generated_structuring(cl_and_vals, dv: bool):
    converter = Converter(detailed_validation=dv)
    cl, vals, kwargs = cl_and_vals
    fn = make_dict_structure_fn(cl, converter, _cattrs_detailed_validation=dv)
    assert fn.overrides == {}

    inst = cl(*vals, **kwargs)

    unstructured = converter.unstructure(inst)

    assert "Hyp" not in repr(unstructured)

    converter.register_structure_hook(cl, fn)

    res = converter.structure(unstructured, cl)

    assert inst == res


@given(
    simple_typed_classes(min_attrs=1) | simple_typed_dataclasses(min_attrs=1), data()
)
def test_renaming(cl_and_vals, data):
    converter = Converter()
    cl, vals, kwargs = cl_and_vals
    attrs = fields(cl)

    to_replace = data.draw(sampled_from(attrs))

    u_fn = make_dict_unstructure_fn(
        cl, converter, **{to_replace.name: override(rename="class")}
    )
    s_fn = make_dict_structure_fn(
        cl, converter, **{to_replace.name: override(rename="class")}
    )
    assert s_fn.overrides == {to_replace.name: override(rename="class")}
    assert u_fn.overrides == {to_replace.name: override(rename="class")}

    converter.register_structure_hook(cl, s_fn)
    converter.register_unstructure_hook(cl, u_fn)

    inst = cl(*vals, **kwargs)

    raw = converter.unstructure(inst)

    assert "class" in raw

    new_inst = converter.structure(raw, cl)

    assert inst == new_inst


def test_renaming_forbid_extra_keys():
    converter = BaseConverter()

    @define
    class A:
        b: int
        c: str

    s_fn = make_dict_structure_fn(
        A, converter, c=override(rename="d"), _cattrs_forbid_extra_keys=True
    )

    converter.register_structure_hook(A, s_fn)

    new_inst = converter.structure({"b": 1, "d": "str"}, A)

    assert new_inst == A(1, "str")

    with pytest.raises(ClassValidationError) as cve:
        converter.structure({"b": 1, "c": "str"}, A)

    assert len(cve.value.exceptions) == 2
    assert isinstance(cve.value.exceptions[0], KeyError)
    assert isinstance(cve.value.exceptions[1], ForbiddenExtraKeysError)
    assert cve.value.exceptions[1].cl is A
    assert cve.value.exceptions[1].extra_fields == {"c"}


def test_omitting(converter: BaseConverter):
    """Omitting works."""

    @define
    class A:
        a: int
        b: int = field(init=False)

    converter.register_unstructure_hook(
        A, make_dict_unstructure_fn(A, converter, b=override(omit=True))
    )

    assert converter.unstructure(A(1)) == {"a": 1}


def test_omitting_none(converter: BaseConverter):
    """Omitting works properly with None."""

    @define
    class A:
        a: int
        b: int = field(init=False)

    converter.register_unstructure_hook(
        A, make_dict_unstructure_fn(A, converter, a=override(), b=override())
    )

    assert converter.unstructure(A(1)) == {"a": 1}

    converter.register_structure_hook(
        A, make_dict_structure_fn(A, converter, a=override(), b=override())
    )

    assert converter.structure({"a": 2}, A).a == 2
    assert not hasattr(converter.structure({"a": 2}, A), "b")


@pytest.mark.parametrize("detailed_validation", [True, False])
def test_omitting_structure(detailed_validation: bool):
    """Omitting fields works with generated structuring functions."""
    converter = BaseConverter()

    @define
    class A:
        a: int
        b: int = field(init=False)
        c: int = 1

    converter.register_structure_hook(
        A,
        make_dict_structure_fn(
            A,
            converter,
            b=override(omit=True),
            c=override(omit=True),
            _cattrs_detailed_validation=detailed_validation,
        ),
    )

    # The 'c' parameter is ignored and the default is used instead.
    structured = converter.structure({"a": 1, "b": 2, "c": 3}, A)
    assert structured.a == 1
    assert structured.c == 1
    assert not hasattr(structured, "b")


@pytest.mark.skipif(not is_py39_plus, reason="literals and annotated are 3.9+")
def test_type_names_with_quotes():
    """Types with quote characters in their reprs should work."""
    from typing import Annotated, Literal, Union

    converter = Converter()

    assert converter.structure({1: 1}, Dict[Annotated[int, "'"], int]) == {1: 1}

    converter.register_structure_hook_func(
        lambda t: t is Union[Literal["a", 2, 3], Literal[4]], lambda v, _: v
    )
    assert converter.structure(
        {2: "a"}, Dict[Union[Literal["a", 2, 3], Literal[4]], str]
    ) == {2: "a"}


def test_overriding_struct_hook(converter: BaseConverter) -> None:
    """Overriding structure hooks works."""
    from math import ceil

    @define
    class A:
        a: int
        b: str

    converter.register_structure_hook(
        A,
        make_dict_structure_fn(
            A,
            converter,
            a=override(struct_hook=lambda v, _: ceil(v)),
            _cattrs_detailed_validation=converter.detailed_validation,
        ),
    )

    assert converter.structure({"a": 0.5, "b": 1}, A) == A(1, "1")


def test_overriding_unstruct_hook(converter: BaseConverter) -> None:
    """Overriding unstructure hooks works."""

    @define
    class A:
        a: int
        b: str

    converter.register_unstructure_hook(
        A,
        make_dict_unstructure_fn(
            A, converter, a=override(unstruct_hook=lambda v: v + 1)
        ),
    )

    assert converter.unstructure(A(1, "")) == {"a": 2, "b": ""}


def test_alias_keys(converter: BaseConverter) -> None:
    """Alias keys work."""

    @define
    class A:
        _a: int
        b: int = field(alias="aliased")
        c: int = field(alias="also_aliased", default=3)
        d: int = field(alias="d_aliased", default=5)

    converter.register_unstructure_hook(
        A,
        make_dict_unstructure_fn(
            A,
            converter,
            _cattrs_use_alias=True,
            c=override(omit=True),
            d=override(rename="d_renamed"),
        ),
    )

    assert converter.unstructure(A(1, 2, 3, 4)) == {
        "a": 1,
        "aliased": 2,
        "d_renamed": 4,
    }

    converter.register_structure_hook(
        A,
        make_dict_structure_fn(
            A,
            converter,
            _cattrs_use_alias=True,
            _cattrs_detailed_validation=converter.detailed_validation,
            c=override(omit=True),
            d=override(rename="d_renamed"),
        ),
    )

    assert converter.structure({"a": 1, "aliased": 2, "d_renamed": 4}, A) == A(
        1, 2, 3, 4
    )


def test_init_false(converter: BaseConverter) -> None:
    """By default init=False keys are ignored."""

    @define
    class A:
        a: int
        b: int = field(init=False)
        _c: int = field(init=False)
        d: int = field(init=False, default=4)

    converter.register_unstructure_hook(A, make_dict_unstructure_fn(A, converter))

    a = A(1)
    a.b = 2
    a._c = 3

    assert converter.unstructure(a) == {"a": 1}

    converter.register_structure_hook(
        A,
        make_dict_structure_fn(
            A, converter, _cattrs_detailed_validation=converter.detailed_validation
        ),
    )

    structured = converter.structure({"a": 1}, A)

    assert not hasattr(structured, "b")
    assert not hasattr(structured, "_c")
    assert structured.d == 4
    assert structured.a == 1


def test_init_false_overridden(converter: BaseConverter) -> None:
    """init=False handling can be overriden."""

    @frozen
    class Inner:
        a: int

    @define
    class A:
        a: int
        b: int = field(init=False)
        _c: int = field(init=False)
        d: Inner = field(init=False)
        e: int = field(init=False, default=4)
        f: Inner = field(init=False, default=Inner(1))

    converter.register_unstructure_hook(
        A, make_dict_unstructure_fn(A, converter, _cattrs_include_init_false=True)
    )

    a = A(1)
    a.b = 2
    a._c = 3
    a.d = Inner(4)

    assert converter.unstructure(a) == {
        "a": 1,
        "b": 2,
        "_c": 3,
        "d": {"a": 4},
        "e": 4,
        "f": {"a": 1},
    }

    converter.register_structure_hook(
        A, make_dict_structure_fn(A, converter, _cattrs_include_init_false=True)
    )

    structured = converter.structure({"a": 1, "b": 2, "_c": 3, "d": {"a": 1}}, A)
    assert structured.b == 2
    assert structured._c == 3
    assert structured.d == Inner(1)
    assert structured.e == 4
    assert structured.f == Inner(1)

    structured = converter.structure(
        {"a": 1, "b": 2, "_c": 3, "d": {"a": 5}, "e": -4, "f": {"a": 2}}, A
    )
    assert structured.b == 2
    assert structured._c == 3
    assert structured.d == Inner(5)
    assert structured.e == -4
    assert structured.f == Inner(2)


def test_init_false_field_override(converter: BaseConverter) -> None:
    """init=False handling can be overriden on a per-field basis."""

    @define
    class A:
        a: int
        b: int = field(init=False)
        _c: int = field(init=False)
        d: int = field(init=False, default=4)

    converter.register_unstructure_hook(
        A,
        make_dict_unstructure_fn(
            A,
            converter,
            b=override(omit=False),
            _c=override(omit=False),
            d=override(omit=False),
        ),
    )

    a = A(1)
    a.b = 2
    a._c = 3

    assert converter.unstructure(a) == {"a": 1, "b": 2, "_c": 3, "d": 4}

    converter.register_structure_hook(
        A,
        make_dict_structure_fn(
            A,
            converter,
            b=override(omit=False),
            _c=override(omit=False),
            d=override(omit=False),
            _cattrs_detailed_validation=converter.detailed_validation,
        ),
    )

    structured = converter.structure({"a": 1, "b": 2, "_c": 3}, A)
    assert structured.b == 2
    assert structured._c == 3
    assert structured.d == 4

    structured = converter.structure({"a": 1, "b": 2, "_c": 3, "d": -4}, A)
    assert structured.b == 2
    assert structured._c == 3
    assert structured.d == -4


def test_init_false_no_structure_hook(converter: BaseConverter):
    """init=False attributes with converters and `prefer_attrs_converters` work."""

    @define
    class A:
        a: int = field(converter=int, init=False)

    converter.register_structure_hook(
        A,
        make_dict_structure_fn(
            A,
            converter,
            _cattrs_prefer_attrib_converters=True,
            _cattrs_include_init_false=True,
        ),
    )

    res = A()
    res.a = 5

    assert converter.structure({"a": "5"}, A) == res


@given(forbid_extra_keys=..., detailed_validation=...)
def test_forbid_extra_keys_from_converter(
    forbid_extra_keys: bool, detailed_validation: bool
):
    """
    `forbid_extra_keys` is taken from the converter by default.
    """
    c = Converter(
        forbid_extra_keys=forbid_extra_keys, detailed_validation=detailed_validation
    )

    @define
    class A:
        a: int

    c.register_structure_hook(A, make_dict_structure_fn(A, c))

    if forbid_extra_keys:
        with pytest.raises((ForbiddenExtraKeysError, ClassValidationError)):
            c.structure({"a": 1, "b": 2}, A)
    else:
        c.structure({"a": 1, "b": 2}, A)


@given(detailed_validation=...)
def test_forbid_extra_keys_from_baseconverter(detailed_validation: bool):
    """
    `forbid_extra_keys` is taken from the converter by default.

    BaseConverter should default to False.
    """
    c = BaseConverter(detailed_validation=detailed_validation)

    @define
    class A:
        a: int

    c.register_structure_hook(A, make_dict_structure_fn(A, c))

    c.structure({"a": 1, "b": 2}, A)


def test_detailed_validation_from_converter(converter: BaseConverter):
    """
    `detailed_validation` is taken from the converter by default.
    """

    @define
    class A:
        a: int

    converter.register_structure_hook(A, make_dict_structure_fn(A, converter))

    if converter.detailed_validation:
        with pytest.raises(ClassValidationError):
            converter.structure({"a": "a"}, A)
    else:
        with pytest.raises(ValueError):
            converter.structure({"a": "a"}, A)


@given(prefer=...)
def test_prefer_converters_from_converter(prefer: bool):
    """
    `prefer_attrs_converters` is taken from the converter by default.
    """

    @define
    class A:
        a: int = field(converter=lambda x: x + 1)

    converter = BaseConverter(prefer_attrib_converters=prefer)
    converter.register_structure_hook(int, lambda x, _: x + 1)
    converter.register_structure_hook(A, make_dict_structure_fn(A, converter))

    if prefer:
        assert converter.structure({"a": 1}, A).a == 2
    else:
        assert converter.structure({"a": 1}, A).a == 3


def test_fields_exception():
    """fields() raises on a non-attrs, non-dataclass class."""
    with pytest.raises(Exception):  # noqa: B017
        fields(int)



"""Tests for generics under PEP 696 (type defaults)."""

from typing import Generic, List

import pytest
from attrs import define, fields
from typing_extensions import TypeVar

from cattrs.errors import StructureHandlerNotFoundError
from cattrs.gen import generate_mapping

T = TypeVar("T")
TD = TypeVar("TD", default=str)


def test_structure_typevar_default(genconverter):
    """Generics with defaulted TypeVars work."""

    @define
    class C(Generic[T]):
        a: T

    c_mapping = generate_mapping(C)
    atype = fields(C).a.type
    assert atype.__name__ not in c_mapping

    with pytest.raises(StructureHandlerNotFoundError):
        # Missing type for generic argument
        genconverter.structure({"a": "1"}, C)

    c_mapping = generate_mapping(C[str])
    atype = fields(C[str]).a.type
    assert c_mapping[atype.__name__] == str

    assert genconverter.structure({"a": "1"}, C[str]) == C("1")

    @define
    class D(Generic[TD]):
        a: TD

    d_mapping = generate_mapping(D)
    atype = fields(D).a.type
    assert d_mapping[atype.__name__] == str

    # Defaults to string
    assert d_mapping[atype.__name__] == str
    assert genconverter.structure({"a": "1"}, D) == D("1")

    # But allows other types
    assert genconverter.structure({"a": "1"}, D[str]) == D("1")
    assert genconverter.structure({"a": 1}, D[int]) == D(1)


def test_unstructure_iterable(genconverter):
    """Unstructuring iterables with defaults works."""
    genconverter.register_unstructure_hook(str, lambda v: v + "_str")

    @define
    class C(Generic[TD]):
        a: List[TD]

    assert genconverter.unstructure(C(["a"])) == {"a": ["a_str"]}
    assert genconverter.unstructure(["a"], List[TD]) == ["a_str"]



"""Tests for the `cattrs.cols` module."""

from immutables import Map

from cattrs import BaseConverter, Converter
from cattrs._compat import AbstractSet, FrozenSet
from cattrs.cols import is_any_set, iterable_unstructure_factory


def test_set_overriding(converter: BaseConverter):
    """Overriding abstract sets by wrapping the default factory works."""

    converter.register_unstructure_hook_factory(
        is_any_set,
        lambda t, c: iterable_unstructure_factory(t, c, unstructure_to=sorted),
    )

    assert converter.unstructure({"c", "b", "a"}, AbstractSet[str]) == ["a", "b", "c"]
    assert converter.unstructure(frozenset(["c", "b", "a"]), FrozenSet[str]) == [
        "a",
        "b",
        "c",
    ]


def test_structuring_immutables_map(genconverter: Converter):
    """This should work due to our new is_mapping predicate."""
    assert genconverter.structure({"a": 1}, Map[str, int]) == Map(a=1)



"""Tests for msgspec functionality."""

from __future__ import annotations

from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Mapping,
    MutableMapping,
    MutableSequence,
    NamedTuple,
    Sequence,
)

from attrs import define
from hypothesis import given
from msgspec import Struct, to_builtins
from pytest import fixture

from cattrs.fns import identity
from cattrs.preconf.json import make_converter as make_json_converter
from cattrs.preconf.msgspec import MsgspecJsonConverter as Conv
from cattrs.preconf.msgspec import make_converter

from ..typed import simple_typed_classes


@define
class A:
    a: int


@define
class B:
    """This class should not be passed through to msgspec."""

    a: Any


@define
class C:
    """This class should not be passed through due to a private attribute."""

    _a: int


class N(NamedTuple):
    a: int


class NA(NamedTuple):
    """A complex namedtuple."""

    a: A


class NC(NamedTuple):
    """A complex namedtuple."""

    a: C


class E(Enum):
    TEST = "test"


@fixture
def converter() -> Conv:
    return make_converter()


def is_passthrough(fn: Callable) -> bool:
    return fn in (identity, to_builtins)


def test_unstructure_passthrough(converter: Conv):
    """Passthrough for simple types works."""
    assert converter.get_unstructure_hook(int) == identity
    assert converter.get_unstructure_hook(float) == identity
    assert converter.get_unstructure_hook(str) == identity
    assert is_passthrough(converter.get_unstructure_hook(bytes))
    assert converter.get_unstructure_hook(None) == identity
    assert is_passthrough(converter.get_unstructure_hook(Literal[1]))
    assert is_passthrough(converter.get_unstructure_hook(E))

    # Any is special-cased, and we cannot know if it'll match
    # the msgspec behavior.
    assert not is_passthrough(converter.get_unstructure_hook(List))
    assert not is_passthrough(converter.get_unstructure_hook(Sequence))
    assert not is_passthrough(converter.get_unstructure_hook(MutableSequence))
    assert not is_passthrough(converter.get_unstructure_hook(List[Any]))
    assert not is_passthrough(converter.get_unstructure_hook(Sequence))
    assert not is_passthrough(converter.get_unstructure_hook(MutableSequence))

    assert is_passthrough(converter.get_unstructure_hook(List[int]))
    assert is_passthrough(converter.get_unstructure_hook(Sequence[int]))
    assert is_passthrough(converter.get_unstructure_hook(MutableSequence[int]))


def test_unstructure_pt_product_types(converter: Conv):
    """Passthrough for product types (attrs, dataclasses...) works."""
    assert is_passthrough(converter.get_unstructure_hook(A))
    assert not is_passthrough(converter.get_unstructure_hook(B))
    assert not is_passthrough(converter.get_unstructure_hook(C))

    assert is_passthrough(converter.get_unstructure_hook(N))
    assert is_passthrough(converter.get_unstructure_hook(NA))
    assert not is_passthrough(converter.get_unstructure_hook(NC))


def test_unstructure_pt_mappings(converter: Conv):
    """Mapping are passed through for unstructuring."""
    assert is_passthrough(converter.get_unstructure_hook(Dict[str, str]))
    assert is_passthrough(converter.get_unstructure_hook(Dict[int, int]))

    assert not is_passthrough(converter.get_unstructure_hook(Dict))
    assert not is_passthrough(converter.get_unstructure_hook(dict))
    assert not is_passthrough(converter.get_unstructure_hook(Dict[int, B]))
    assert not is_passthrough(converter.get_unstructure_hook(Mapping))
    assert not is_passthrough(converter.get_unstructure_hook(MutableMapping))

    assert is_passthrough(converter.get_unstructure_hook(Dict[int, A]))
    assert is_passthrough(converter.get_unstructure_hook(Mapping[int, int]))
    assert is_passthrough(converter.get_unstructure_hook(MutableMapping[int, int]))


def test_dump_hook(converter: Conv):
    """Passthrough for dump hooks works."""
    assert converter.get_dumps_hook(A) == converter.encoder.encode
    assert converter.get_dumps_hook(Dict[str, str]) == converter.encoder.encode

    # msgspec cannot handle these, so cattrs does.
    assert converter.get_dumps_hook(B) == converter.dumps


def test_get_loads_hook(converter: Conv):
    """`Converter.get_loads_hook` works."""
    hook = converter.get_loads_hook(A)
    assert hook(b'{"a": 1}') == A(1)


def test_basic_structs(converter: Conv):
    """Handling msgspec structs works."""

    class B(Struct):
        b: int

    assert converter.unstructure(B(1)) == {"b": 1}

    assert converter.structure({"b": 1}, B) == B(1)


@given(simple_typed_classes(text_codec="ascii", allow_infinity=False, allow_nan=False))
def test_simple_classes(cls_and_vals):
    cl, posargs, kwargs = cls_and_vals

    msgspec = make_converter()
    json = make_json_converter()

    inst = cl(*posargs, **kwargs)

    rebuilt_msgspec = msgspec.loads(msgspec.dumps(inst), cl)
    rebuilt_json = json.loads(json.dumps(inst), cl)

    assert rebuilt_msgspec == rebuilt_json



"""Tests for tuples under PEP 563 (stringified annotations)."""

from __future__ import annotations

from typing import NamedTuple

from cattrs import Converter
from cattrs.cols import (
    namedtuple_dict_structure_factory,
    namedtuple_dict_unstructure_factory,
)


class NT(NamedTuple):
    a: int


def test_simple_dict_nametuples(genconverter: Converter):
    """Namedtuples can be un/structured to/from dicts."""

    class Test(NamedTuple):
        a: int
        b: str = "test"

    genconverter.register_unstructure_hook_factory(
        lambda t: t is Test, namedtuple_dict_unstructure_factory
    )
    genconverter.register_structure_hook_factory(
        lambda t: t is Test, namedtuple_dict_structure_factory
    )

    assert genconverter.unstructure(Test(1)) == {"a": 1, "b": "test"}
    assert genconverter.structure({"a": 1, "b": "2"}, Test) == Test(1, "2")

    # Defaults work.
    assert genconverter.structure({"a": 1}, Test) == Test(1, "test")



"""Tests for collections in `cattrs.gen`."""

from typing import Generic, Mapping, NewType, Tuple, TypeVar

from cattrs import Converter
from cattrs.gen import make_hetero_tuple_unstructure_fn, make_mapping_structure_fn


def test_structuring_mappings(genconverter: Converter):
    """The `key_type` parameter works for generics with 1 type variable."""
    T = TypeVar("T")

    class MyMapping(Generic[T], Mapping[str, T]):
        pass

    def key_hook(value, _):
        return f"{value}1"

    Key = NewType("Key", str)

    genconverter.register_structure_hook(Key, key_hook)

    fn = make_mapping_structure_fn(MyMapping[int], genconverter, key_type=Key)

    assert fn({"a": 1}, MyMapping[int]) == {"a1": 1}


def test_unstructure_hetero_tuple_to_tuple(genconverter: Converter):
    """`make_hetero_tuple_unstructure_fn` works when unstructuring to tuple."""
    fn = make_hetero_tuple_unstructure_fn(Tuple[int, str, int], genconverter, tuple)

    assert fn((1, "1", 2)) == (1, "1", 2)



"""Tests for PEP 695 (Type Parameter Syntax)."""

from dataclasses import dataclass

import pytest
from attrs import define

from cattrs import BaseConverter, Converter

from ._compat import is_py312_plus


@pytest.mark.skipif(not is_py312_plus, reason="3.12+ syntax")
def test_simple_generic_roundtrip(converter: BaseConverter):
    """PEP 695 attrs generics work."""

    @define
    class A[T]:
        a: T

    assert converter.structure({"a": "1"}, A[int]) == A(1)
    assert converter.unstructure(A(1)) == {"a": 1}

    if isinstance(converter, Converter):
        # Only supported on a Converter
        assert converter.unstructure(A(1), A[int]) == {"a": 1}


@pytest.mark.skipif(not is_py312_plus, reason="3.12+ syntax")
def test_simple_generic_roundtrip_dc(converter: BaseConverter):
    """PEP 695 dataclass generics work."""

    @dataclass
    class A[T]:
        a: T

    assert converter.structure({"a": "1"}, A[int]) == A(1)
    assert converter.unstructure(A(1)) == {"a": 1}

    if isinstance(converter, Converter):
        # Only supported on a Converter
        assert converter.unstructure(A(1), A[int]) == {"a": 1}


def test_type_aliases(converter: BaseConverter):
    """PEP 695 type aliases work."""
    type my_int = int

    assert converter.structure("42", my_int) == 42
    assert converter.unstructure(42, my_int) == 42

    type my_other_int = int

    # Manual hooks should work.

    converter.register_structure_hook_func(
        lambda t: t is my_other_int, lambda v, _: v + 10
    )
    converter.register_unstructure_hook_func(
        lambda t: t is my_other_int, lambda v: v - 20
    )

    assert converter.structure(1, my_other_int) == 11
    assert converter.unstructure(100, my_other_int) == 80


def test_type_aliases_overwrite_base_hooks(converter: BaseConverter):
    """Overwriting base hooks should affect type aliases."""
    converter.register_structure_hook(int, lambda v, _: v + 10)
    converter.register_unstructure_hook(int, lambda v: v - 20)

    type my_int = int

    assert converter.structure(1, my_int) == 11
    assert converter.unstructure(100, my_int) == 80


def test_type_alias_with_children(converter: BaseConverter):
    """A type alias that chains to a hook that requires the type parameter works."""

    class TestClass:
        pass

    def structure_testclass(val, type):
        assert type is TestClass
        return TestClass

    converter.register_structure_hook(TestClass, structure_testclass)

    type TestAlias = TestClass
    assert converter.structure(None, TestAlias) is TestClass



"""Tests for handling `typing.Any`."""

from typing import Any, Dict, Optional

from attrs import define
from typing_extensions import Any as ExtendedAny


@define
class A:
    pass


def test_unstructuring_dict_of_any(converter):
    """Dicts with Any values should use runtime dispatch for their values."""
    assert converter.unstructure({"a": A()}, Dict[str, Any]) == {"a": {}}


def test_unstructuring_any(converter):
    """`Any` should use runtime dispatch."""

    assert converter.unstructure(A(), Any) == {}


def test_unstructure_optional_any(converter):
    """Unstructuring `Optional[Any]` should use the runtime value."""

    assert converter.unstructure(A(), Optional[Any]) == {}


def test_extended_any(converter):
    """`typing_extensions.Any` works."""

    assert converter.unstructure(A(), unstructure_as=ExtendedAny) == {}

    d = {}
    assert converter.structure(d, ExtendedAny) is d
